#include "bindGroups.h"
#include "buffer.h"

#include "ctre.hpp"

#include<cassert>
#include<ranges>


namespace ranges = std::ranges;

bool contains(const std::string_view text, const std::string_view match) {
	return text.find(match) != std::string_view::npos;
}

[[nodiscard]] wgpu::BufferBindingType parseBufferType(const std::string_view type,const std::string_view access) {
	using enum wgpu::BufferBindingType;
	if (type == "uniform") {
		return Uniform;
	}
	assert(type == "storage");
	if (access == "read_write") {
		return Storage;
	}
	assert(access == "read");
	return ReadOnlyStorage;
}

[[nodiscard]] wgpu::TextureSampleType parseSampleType(std::string_view type) {
	using enum  wgpu::TextureSampleType;
	if (contains(type, "f32")) return Float;
	if (contains(type, "i32")) return Sint;
	if (contains(type, "u32")) return Uint;
	//missing UnfilterableFloat
	assert(false);
	return Float;
}

[[nodiscard]] wgpu::TextureViewDimension parseTextureDimension(const std::string_view dimension) {
	using enum wgpu::TextureViewDimension;
	if (contains(dimension, "1d")) return e1D;
	if (contains(dimension, "2d_array")) return e2DArray;
	if (contains(dimension, "2d")) return e2D;
	if (contains(dimension, "3d")) return e3D;
	if (contains(dimension, "cube_array")) return CubeArray;
	if (contains(dimension, "cube")) return Cube;
	assert(false);
	return e2D;
}

[[nodiscard]] wgpu::TextureFormat parseTextureFormat(const std::string_view format) {
	using enum wgpu::TextureFormat;
	constexpr auto strings = std::to_array<std::string_view>({ "rgba8unorm", "rgba8snorm", "rgba8uint", "rgba8sint", "rgba16uint", "rgba16sint", "rgba16float", "r32uint", "r32sint", "r32float", "rg32uint", "rg32sint", "rg32float", "rgba32uint", "rgba32sint", "rgba32float" });
	constexpr std::array formats{ RGBA8Unorm, RGBA8Snorm, RGBA8Uint, RGBA8Sint, RGBA16Uint, RGBA16Sint, RGBA16Float, R32Uint, R32Sint, R32Float, RG32Uint, RG32Sint, RG32Float, RGBA32Uint, RGBA32Sint, RGBA32Float };
	static_assert(strings.size() == formats.size());
	const auto pos = ranges::find(strings, format) - strings.begin();
	assert(pos < int64_t{ formats.size() });
	return formats[pos];
}

std::vector<wgpu::BindGroupLayout> createBindGroupLayouts(const wgpu::Device& device, const Shader& shader, wgpu::ShaderStage shaderStage) {
	constexpr auto not_space = [](const char c) {return !std::isspace(c); };
	auto view_filtered = std::views::filter(shader.source, not_space);
	const std::string filtered = { view_filtered.begin(), view_filtered.end() };


	std::vector<std::vector<wgpu::BindGroupLayoutEntry>> entries;
	for (auto& [whole, group, binding, variable] : ctre::range<R"(\[\[group\((\d*)\),binding\((\d*)\)\]\]var([^;]*))">(filtered)) {
		wgpu::BindGroupLayoutEntry entry{
			.binding = std::stoul(binding.to_string()),
			.visibility = shaderStage,
		};

		if (auto [match, type, access] = ctre::match<R"(<([^,>]*),?([^>]*)>.*)">(variable); match) {
			entry.buffer.type = parseBufferType(type, access);
		}
		else if (ctre::match<R"([^:]*:sampler)">(variable)) {
			entry.sampler.type = wgpu::SamplerBindingType::Filtering;
		}
		else if (auto [match, dimension, format, access] = ctre::match<R"([^:]*:texture_storage_([^<]*)<([^>,]*),([^>]*)>)">(variable); match) {
			assert(access == "write");
			entry.storageTexture.access = wgpu::StorageTextureAccess::WriteOnly;
			entry.storageTexture.format = parseTextureFormat(format.to_view());
			entry.storageTexture.viewDimension = parseTextureDimension(dimension);
		}
		else if (auto [match, kind, type] = ctre::match<R"([^:]*:texture_([^<]*)<([^>]*)>)">(variable); match) {
			entry.texture.sampleType = contains(kind, "depth") ? wgpu::TextureSampleType::Depth : parseSampleType(type);
			entry.texture.viewDimension = parseTextureDimension(kind);
			entry.texture.multisampled = contains(kind, "multisampled");
		}
		else {
			assert(false);
		}

		const auto group_id = std::stoul(group.to_string());
		if (group_id >= entries.size()) {
			entries.resize(size_t{ group_id } + 1);
		}
		entries[group_id].push_back(entry);
	}

	return createBindGroupLayouts(device, entries);
}

std::vector<wgpu::BindGroupLayout> createBindGroupLayouts(const wgpu::Device& device, const std::vector<std::vector<wgpu::BindGroupLayoutEntry>>& entries) {
	
	std::vector<wgpu::BindGroupLayout> result(entries.size());
	for (unsigned i = 0; i < entries.size(); i++) {
		const wgpu::BindGroupLayoutDescriptor bindGroupLayoutDesc{
			.entryCount = static_cast<uint32_t>(entries[i].size()),
			.entries = entries[i].data()
		};
		result[i] = device.CreateBindGroupLayout(&bindGroupLayoutDesc);
	}
	return result;
}
