#include "buffer.h"
#include "context.h"


void BufferSubrange::getData(const std::span<std::byte> data) const {
	Buffer::stagingBuffers->gpuToCpu.copy(*this, data);
}

BufferSubrange::BufferSubrange(Buffer& buffer, const uint64_t offset, const uint64_t size) :
	_buffer{ buffer }, _offset{ offset }, _size{ size }
{
	assert(offset + size <= buffer.desc.size);
}


BufferSubrange::BufferSubrange(Buffer& buffer) :
	BufferSubrange{ buffer, 0, buffer.desc.size } { }


void StagingBuffer::reserve(const size_t size) {
	if (size > desc.size) {
		desc.size = size;
		buffer = context.device.CreateBuffer(&desc);
	}
}


void CPU_to_GPU_StagingBuffer::copy_to(BufferSubrange dst) {
	context.encoder.CopyBufferToBuffer(buffer, 0, dst.buffer(), dst.offset(), dst.size());
}

void CPU_to_GPU_StagingBuffer::copy(std::span<const std::byte> src, BufferSubrange dst) {
	context.flushCommands();
	reserve(src.size());
	mapAsync();
	assert(src.size() == dst.size());
	auto* buffer_ptr = getMapped();
	std::memcpy(buffer_ptr, src.data(), src.size());
	unmap();
	copy_to(dst);
}


void GPU_to_CPU_StagingBuffer::copy_from(const BufferSubrange& src) {
	reserve(src.size());
	context.encoder.CopyBufferToBuffer(src.buffer(), src.offset(), buffer, 0, src.size());
}


void GPU_to_CPU_StagingBuffer::copy(BufferSubrange src, std::span<std::byte> dst) {
	assert(src.size() == dst.size());
	copy_from(src);
	context.flushCommandsAndWait();
	mapAsync();
	auto* buffer_ptr = getMapped();
	std::memcpy(dst.data(), buffer_ptr, dst.size());
	unmap();
}


Buffer::Buffer(const Context& context, const wgpu::BufferDescriptor& bufferDesc, const std::byte* data)
	: context{ context }, buffer{ createBuffer(context.device, bufferDesc) }, desc{ bufferDesc }
{
	if (!stagingBuffers) {
		stagingBuffers = new StagingBufferPair{ { context }, { context } };
	}
	if (data) {
		stagingBuffers->cpuToGpu.copy({ data, bufferDesc.size }, *this);
	}
}


void Buffer::copyToRenderImage() {
	const wgpu::TextureDataLayout bufferLayout{
		.bytesPerRow = context.imageSize.width * 4,
		.rowsPerImage = context.imageSize.height
	};
	const wgpu::ImageCopyBuffer srcBuffer{ .layout = bufferLayout, .buffer = buffer };
	const wgpu::ImageCopyTexture dstTexture{ .texture = context.renderImage, .origin{} };

	context.encoder.CopyBufferToTexture(&srcBuffer, &dstTexture, &context.imageSize);
}

void setMapped(const WGPUBufferMapAsyncStatus status, void* mappedFlag) {
	if (status != WGPUBufferMapAsyncStatus_Success) {
		assert(false);
	}
	auto* mapped = reinterpret_cast<std::atomic<StagingBuffer::MappingStatus>*>(mappedFlag);
	assert(mapped);
	*mapped = StagingBuffer::MappingStatus::Mapped;
}
