#pragma once

#include "context.h"
#include "common.h"

#include <atomic>
#include <cassert>
#include <ranges>
#include <iostream>


class Buffer;

class BufferSubrange {
	Buffer& _buffer;
	uint64_t _offset;
	uint64_t _size;
public:
	Buffer& buffer() const {
		return _buffer;
	}
	uint64_t offset() const {
		return _offset;
	}
	uint64_t size() const {
		return _size;
	}
	std::pair<BufferSubrange, uint64_t> subrange(uint64_t offset, uint64_t size) {
		return { BufferSubrange{_buffer, offset, size}, roundUp(offset + size, bufferSubrangeAllignment) };
	}
	void getData(std::span<std::byte> data) const;
	void getData(std::byte* data, size_t size) const {
		getData(std::span{ data, size });
	}
	BufferSubrange(Buffer& buffer, uint64_t offset, uint64_t size);
	BufferSubrange(Buffer& buffer);
};

void setMapped(WGPUBufferMapAsyncStatus status, void* mappedFlag);

class StagingBuffer {
public:
	wgpu::BufferDescriptor desc;
	const Context& context;
	wgpu::Buffer buffer;
	
	enum class MappingStatus{ Unmapped, InProgress, Mapped };
	std::atomic<MappingStatus> mapped;
	byte* data;

	StagingBuffer(const Context& context, wgpu::BufferUsage usage, bool mappedAtCreation) :
		context(context), 
		desc{ .usage = usage, .size = 0, .mappedAtCreation = mappedAtCreation},
		mapped{ mappedAtCreation ? MappingStatus::Mapped : MappingStatus::Unmapped } { }

	void reserve(size_t size);

	void unmap() {
		auto oldVal = MappingStatus::Mapped;
		if (mapped.compare_exchange_strong(oldVal, MappingStatus::Unmapped)) {
			buffer.Unmap();
		}
	}

protected:
	template<wgpu::MapMode mapMode>
	void mapAsyncTemplate() {
		auto oldVal = MappingStatus::Unmapped;
		if (mapped.compare_exchange_strong(oldVal, MappingStatus::InProgress)) {
			buffer.MapAsync(mapMode, 0, desc.size, &setMapped, &this->mapped);
			context.device.Tick();
		}
	}

	~StagingBuffer() = default;
};

class CPU_to_GPU_StagingBuffer : public StagingBuffer {
	std::byte* buffer_ptr = nullptr;

public:
	CPU_to_GPU_StagingBuffer(const Context& context)
		: StagingBuffer{ context, wgpu::BufferUsage::MapWrite | wgpu::BufferUsage::CopySrc, true } { }

	void mapAsync() {
		mapAsyncTemplate<wgpu::MapMode::Write>();
	}

	std::byte* getMapped() {
		assert(mapped != MappingStatus::Unmapped);
		while (mapped == MappingStatus::InProgress) {
			context.device.Tick();
		}
		return asWritableBytes(buffer.GetMappedRange(0, desc.size));
	}

	void copy_to(BufferSubrange dst);

	void copy(std::span<const std::byte> src, BufferSubrange dest);
};

class GPU_to_CPU_StagingBuffer : public StagingBuffer {
public:
	const std::byte* buffer_ptr = nullptr;

	GPU_to_CPU_StagingBuffer(const Context& context)
		: StagingBuffer{ context, wgpu::BufferUsage::MapRead | wgpu::BufferUsage::CopyDst, false } { }

	void mapAsync() {
		mapAsyncTemplate<wgpu::MapMode::Read>();
	}

	const std::byte* getMapped(){
		assert(mapped != MappingStatus::Unmapped);
		while (mapped == MappingStatus::InProgress) {
			context.device.Tick();
		}
		auto* mapped = buffer.GetConstMappedRange(0, desc.size);
		auto* result = asBytes(mapped);
		assert(result);
		return result;
	}

	void copy_from(const BufferSubrange& src);

	void copy(BufferSubrange src, std::span<std::byte> dst);
};

class Buffer {
public:
	struct StagingBufferPair {
		CPU_to_GPU_StagingBuffer cpuToGpu;
		GPU_to_CPU_StagingBuffer gpuToCpu;
	};
	static inline StagingBufferPair* stagingBuffers = nullptr;

	const Context& context;
	wgpu::Buffer buffer;
	wgpu::BufferDescriptor desc;

	Buffer(const Context& context,
		uint64_t size, wgpu::BufferUsage usage, bool mappedAtCreation = false,
		const std::byte* data = nullptr)
		: Buffer {
			context,
			wgpu::BufferDescriptor{.usage = usage, .size = size, .mappedAtCreation = mappedAtCreation },
			data
		} { }

	Buffer(const Context& context, const wgpu::BufferDescriptor& bufferDesc, const std::byte* = nullptr);

	void setData(std::span<const std::byte> data) {
		stagingBuffers->cpuToGpu.copy(data, *this);
	}

	void setData(const std::byte* data, size_t size) {
		setData(std::span{ data, size });
	}

	void getData(const std::span<std::byte> data) {
		stagingBuffers->gpuToCpu.copy(*this, data);
	}

	void getData(std::byte* data, size_t size) {
		getData(std::span{ data, size });
	}

	operator const wgpu::Buffer&() const {
		return buffer;
	}

	operator wgpu::Buffer& () {
		return buffer;
	}

	BufferSubrange subrange(uint64_t offset, uint64_t size) {
		return BufferSubrange(*this, offset, size);
	}

	void copyToRenderImage();
private:
	static wgpu::Buffer createBuffer(const wgpu::Device& device, const wgpu::BufferDescriptor& desc) {
		return device.CreateBuffer(&desc);
	}
};

template<typename T>
inline void printRange(const BufferSubrange& range) {
	range.buffer().context.flushCommandsAndWait();
	
	std::vector<T> data(range.size() / sizeof(T));
	range.getData(std::as_writable_bytes(std::span{ data }));

	for (int i = 0; i < data.size(); i++) {
		std::cout << data[i] << " ";
	}
	std::cout << std::endl;
}




