#include "context.h"
#include "buffer.h"
#include "bindGroups.h"
#include "std_common.h"
#include "kernel.h"

#include <GLFW/glfw3.h>
#define GLFW_EXPOSE_NATIVE_WIN32
#include <GLFW/glfw3native.h>

void onMappingDone(WGPUBufferMapAsyncStatus status, void* flagVoidPtr) {
	assert(status == WGPUBufferMapAsyncStatus_Success);
	auto flag = reinterpret_cast<std::atomic_flag*>(flagVoidPtr);
	flag->test_and_set();
}

void onQueueDone(WGPUQueueWorkDoneStatus status, void* flagVoidPtr) {
	assert(status == WGPUQueueWorkDoneStatus_Success);
	auto flag = reinterpret_cast<std::atomic_flag*>(flagVoidPtr);
	flag->test_and_set();
}

Shader::Shader(const wgpu::Device& device, std::filesystem::path filePath) {
	std::ifstream file{ filePath, std::ios::binary };
	if (!file.is_open()) {
		throw std::runtime_error{ "Failed to open file:"s + filePath.string() };
	};
	auto size = std::filesystem::file_size(filePath);
	source.resize(size, '\0');
	file.read(source.data(), static_cast<std::streamsize>(size));
	if (!file.good())
		throw std::runtime_error{ "Error reading from file:"s + filePath.string() };

	wgpu::ShaderModuleWGSLDescriptor wgslShaderDesc;
	wgslShaderDesc.source = source.data();
	//std::cout << shader_code.data() << std::endl;
	wgpu::ShaderModuleDescriptor shaderModuleDescriptor{ .nextInChain = &wgslShaderDesc };
	compiled = device.CreateShaderModule(&shaderModuleDescriptor);
}

namespace {
	struct WindowSize {
		float width;
		float height;
	};

	[[nodiscard]] wgpu::Device createDevice(dawn_native::Instance& dawn_instance) {
		dawn_instance.DiscoverDefaultAdapters();

		const auto adapters = dawn_instance.GetAdapters();
		dawn_native::Adapter adapter;
		for (auto& a : adapters) {
			wgpu::AdapterProperties properties;
			a.GetProperties(&properties);
			//std::cout << properties.name << " " << properties.driverDescription << std::endl;
			if (properties.backendType == wgpu::BackendType::D3D12
				&& properties.adapterType == wgpu::AdapterType::DiscreteGPU)
			{
				adapter = a;
			}
		}
		wgpu::AdapterProperties properties;
		adapter.GetProperties(&properties);
		std::cout << "Chosen adapter:" << properties.name << ", driver: " << properties.driverDescription << std::endl;
		assert(adapter);
		return adapter.CreateDevice();
	}

	[[nodiscard]] wgpu::Surface createSurface(GLFWwindow* window, wgpu::Instance instance) {
		HINSTANCE hinstance = GetModuleHandle(nullptr);

		wgpu::SurfaceDescriptorFromWindowsHWND hwndSurfaceDesc;
		hwndSurfaceDesc.hwnd = glfwGetWin32Window(window);
		hwndSurfaceDesc.hinstance = hinstance;

		wgpu::SurfaceDescriptor surfaceDesc{
			.nextInChain = &hwndSurfaceDesc
		};

		return instance.CreateSurface(&surfaceDesc);
	}

	[[nodiscard]] wgpu::SwapChain createSwapchain(wgpu::Extent3D size,
		const wgpu::Device& device, const wgpu::Surface& surface)
	{
		wgpu::SwapChainDescriptor swapchainDesc{
			.usage = wgpu::TextureUsage::RenderAttachment,
			.format = wgpu::TextureFormat::BGRA8Unorm,
			.width = size.width,
			.height = size.height,
			.presentMode = wgpu::PresentMode::Fifo
		};
		return device.CreateSwapChain(surface, &swapchainDesc);
	}

	static [[nodiscard]] wgpu::RenderPipeline createRenderPipeline(
		const wgpu::Device& device, wgpu::BindGroupLayout bindGroupLayout,
		Shader vertexShader, Shader fragmentShader)
	{
		wgpu::PipelineLayoutDescriptor pipelineLayoutDesc{
			.bindGroupLayoutCount = 1,
			.bindGroupLayouts = &bindGroupLayout
		};
		auto pipelineLayout = device.CreatePipelineLayout(&pipelineLayoutDesc);

		wgpu::VertexState vertex{
				.module = vertexShader,
				.entryPoint = "main",
				.bufferCount = 0,
				.buffers = nullptr,
		};

		wgpu::ColorTargetState colorTarget{ .format = wgpu::TextureFormat::BGRA8Unorm };

		wgpu::FragmentState fragment{
				.module = fragmentShader,
				.entryPoint = "main",
				.targetCount = 1,
				.targets = &colorTarget
		};

		wgpu::RenderPipelineDescriptor renderPipelineDescriptor{
			.layout = pipelineLayout,
			.vertex = vertex,
			.primitive = {},
			.multisample = {},
			.fragment = &fragment
		};
		return device.CreateRenderPipeline(&renderPipelineDescriptor);
	}
}

void Context::initGLFW() {
	if (!glfwInit()) {
		throw std::runtime_error{ "GLFW cannot be initialised." };
	}
	glfwInitialised = true;
	auto errorCallback = [](int error, const char* description) {
		std::cout << "GLFW error: " << description << '\n';
	};
	glfwSetErrorCallback(errorCallback);

	glfwWindowHint(GLFW_CLIENT_API, GLFW_NO_API);
	window = glfwCreateWindow(960, 540, "WebGPU", NULL, NULL);
	if (!window) {
		throw std::runtime_error{ "Window cannot be created." };
	}

	glfwSetWindowUserPointer(window, this);

	auto keycallback = [](GLFWwindow* window, int key, int scancode, int action, int mods) {
		if ((key == GLFW_KEY_ESCAPE || key == GLFW_KEY_Q) && action == GLFW_PRESS) {
			glfwSetWindowShouldClose(window, GLFW_TRUE);
		}
	};
	glfwSetKeyCallback(window, keycallback);
}

Context::Context(bool createWindow){
	dawnProcSetProcs(&dawn_native::GetProcs());
	dawn_native::Instance dawnInstance;
	device = createDevice(dawnInstance);
	assert(device);


	auto errorCallback = [](WGPUErrorType type, const char* msg, void* data) {
		std::cout << "WebGPU Error: " << msg << std::endl;
	};
	device.SetUncapturedErrorCallback(errorCallback, nullptr);
	
	auto deviceLostCallback = [](WGPUDeviceLostReason reason, char const* msg, void* data) {
		std::cout << "WebGPU Error: " << msg << std::endl;
	};
	device.SetDeviceLostCallback(deviceLostCallback, nullptr);

	queue = device.GetQueue();
	encoder = device.CreateCommandEncoder();

	if (createWindow) {
		initGLFW();
		surface = createSurface(window, dawnInstance.Get());

		int width = 0;
		int height = 0;
		glfwGetFramebufferSize(window, &width, &height);
		imageSize = { static_cast<uint32_t>(width), static_cast<uint32_t>(height) };
		swapchain = createSwapchain(imageSize, device, surface);

		using wgpu::TextureUsage;
		wgpu::TextureDescriptor renderImageDesc{
			.usage = TextureUsage::CopyDst | TextureUsage::StorageBinding | TextureUsage::TextureBinding,
			.size = imageSize,
			.format = wgpu::TextureFormat::RGBA8Unorm,
		};
		renderImage = device.CreateTexture(&renderImageDesc);

		vertexShader = Shader(device, "shaders/vert.wgsl");
		fragmentShader = Shader(device, "shaders/frag.wgsl");

		auto layouts = createBindGroupLayouts(device, fragmentShader, wgpu::ShaderStage::Fragment);
		assert(layouts.size() == 1);
		renderBindGroupLayout = layouts[0];

		vertexShader.discardSource();
		fragmentShader.discardSource();
		renderPipeline = createRenderPipeline(device, renderBindGroupLayout, vertexShader, fragmentShader);

		wgpu::SamplerDescriptor samplerDescriptor{
			.addressModeU = wgpu::AddressMode::Repeat,
			.addressModeV = wgpu::AddressMode::Repeat,
			.addressModeW = wgpu::AddressMode::Repeat,
			.magFilter = wgpu::FilterMode::Linear,
			.minFilter = wgpu::FilterMode::Linear
		};
		renderSampler = device.CreateSampler(&samplerDescriptor);

		wgpu::BufferDescriptor windowSizeBufferDesc{
			.usage = wgpu::BufferUsage::Uniform | wgpu::BufferUsage::CopyDst,
			.size = sizeof(WindowSize),
			.mappedAtCreation = false,
		};
		windowSizeUniform = std::make_unique<Buffer>(*this, sizeof(WindowSize), wgpu::BufferUsage::Uniform | wgpu::BufferUsage::CopyDst);
	}
}

Context::~Context(){
	if (window) {
		glfwDestroyWindow(window);
	}
	if (glfwInitialised) {
		glfwTerminate();
	}
}

bool Context::shouldClose() {
	return glfwWindowShouldClose(window);
}

void Context::pollEvents() {
	glfwPollEvents();
}

void Context::drawRenderTexture() {
	int width, height;
	glfwGetWindowSize(window, &width, &height);
	WindowSize windowSize{ static_cast<float>(width), static_cast<float>(height) };
	encoder.WriteBuffer(windowSizeUniform->buffer, 0, reinterpret_cast<uint8_t*>(&windowSize), sizeof(WindowSize));
	auto bindGroup = createBindGroup(device, renderBindGroupLayout, renderSampler, renderImage.CreateView(), *windowSizeUniform);
	wgpu::RenderPassColorAttachment renderPassColorAttachment{
		.view = swapchain.GetCurrentTextureView(),
		.loadOp = wgpu::LoadOp::Clear,
		.storeOp = wgpu::StoreOp::Store,
		.clearColor = wgpu::Color{1.0, 0.0, 0.0, 1.0}
	};

	wgpu::RenderPassDescriptor renderPassDesc{
		.colorAttachmentCount = 1,
		.colorAttachments = &renderPassColorAttachment
	};

	auto renderPassEncoder = encoder.BeginRenderPass(&renderPassDesc);
	renderPassEncoder.SetPipeline(renderPipeline);
	renderPassEncoder.SetBindGroup(0, bindGroup);
	renderPassEncoder.Draw(6);
	renderPassEncoder.EndPass();

	flushCommands();
	swapchain.Present();
}

void Context::flushCommands() const {
	auto commandBuffer = encoder.Finish();
	queue.Submit(1, &commandBuffer);
	encoder = device.CreateCommandEncoder();
}
