#pragma once

#include "context.h"
#include <vector>
#include "bindGroups.h"

namespace {
	[[nodiscard]] wgpu::ComputePipeline createComputePipeline(const wgpu::Device& device, const std::vector<wgpu::BindGroupLayout>& bindGroupLayouts,
		const wgpu::ShaderModule& computeShader)
	{
		std::vector<wgpu::BindGroupLayout> layouts;

		wgpu::PipelineLayoutDescriptor pipelineLayoutDesc{
			.bindGroupLayoutCount = static_cast<uint32_t>(bindGroupLayouts.size()),
			.bindGroupLayouts = bindGroupLayouts.data()
		};
		auto pipelineLayout = device.CreatePipelineLayout(&pipelineLayoutDesc);

		wgpu::ProgrammableStageDescriptor compute{
			.module = computeShader,
			.entryPoint = "main"
		};

		wgpu::ComputePipelineDescriptor computePipelineDescriptor{
			.layout = pipelineLayout,
			.compute = compute
		};
		return device.CreateComputePipeline(&computePipelineDescriptor);
	}
}


class Kernel {

	const Context& _context;
	const Shader _shader;
	std::vector<wgpu::BindGroupLayout> _layouts;
	std::vector<wgpu::BindGroup> _bindGroups;
	wgpu::ComputePipeline _pipeline;
public:
	Kernel(const Context& context, const std::filesystem::path& path)
		: _context{ context }, _shader{ context.device, path }
	{
		_layouts = createBindGroupLayouts(context.device, _shader);
		_bindGroups.resize(_layouts.size());
		_pipeline = createComputePipeline(context.device, _layouts, _shader.compiled);
	}

	template<typename ...T>
	void setBindGroup(unsigned pos, const T&... args) {
		setBindGroup(pos, createBindGroup(_context.device, _layouts[pos], args...));
	}

	void setBindGroup(unsigned pos, wgpu::BindGroup&& bindGroup) {
		_bindGroups[pos] = std::move(bindGroup);
	}

	void dispatch(int x, int y = 1, int z = 1) {
		dispatch(_context.encoder, x, y, z);
		_context.flushCommands();
	}

	void dispatch(wgpu::CommandEncoder encoder, int x, int y = 1, int z = 1) {
		auto computePass = encoder.BeginComputePass();
		computePass.SetPipeline(_pipeline);
		for (unsigned i = 0; i < _bindGroups.size(); i++) {
			if (_bindGroups[i] != nullptr) {
				computePass.SetBindGroup(i, _bindGroups[i]);
			}
		}
		computePass.Dispatch(x, y, z);
		computePass.EndPass();
	}
};