﻿#include "context.h"
#include "common.h"
#include "kernel.h"
#include "std_common.h"
#include "bindGroups.h"
#include "radix.h"


#include "ctre.hpp"

#define STB_IMAGE_IMPLEMENTATION
#include <stb_image.h>

#include <concepts>
#include <algorithm>
#include <array>
#include <chrono>
#include <iostream>
#include <random>
#include <ranges>
#include <span>
#include <vector>
#include <ios>
#include <format>

namespace chrono = std::chrono;
namespace ranges = std::ranges;
namespace views = std::views;

std::mt19937 mt{ 2021 };/*std::random_device{}()*/
constexpr unsigned numKeys = 1024u * 1024u * 64u;

const std::filesystem::path kernelsPath{ R"(D:\Programming_Files\Visual Studio 2022\PV160\kernels\wgsl)" };

namespace {
	struct uniformBuffer {
		uint32_t image_width;
		uint32_t image_height;
	};

	struct PushConstants {
		uint32_t g_num_keys;
		uint32_t g_bit_shift;
	};

	inline std::vector<uint32_t> generate_data(size_t size) {
		static std::uniform_int_distribution<uint32_t> distribution(0u, 1000u);
		static std::default_random_engine                generator;

		std::vector<uint32_t> data(size);
		std::generate(data.begin(), data.end(), []() { return distribution(generator); });
		return data;
    }

	template<typename Arg>
	void print(Arg arg){
		std::cout << std::format("{}", arg);
	}

	template<typename...Args> requires (sizeof...(Args) > 0)
	void print(std::string_view fmt, Args...args){
		std::cout << std::format(fmt, args...);
	}

	void println(){
		std::cout << '\n';
	}

	template<typename Arg>
	void println(Arg arg){
		std::cout << std::format("{}", arg) << std::endl;
	}

	template<typename...Args> requires (sizeof...(Args) > 0)
	void println(std::string_view fmt, Args...args) {
		std::cout << std::format(fmt, args...) << std::endl;
	}

	struct Timer{
		chrono::steady_clock::time_point start = chrono::steady_clock::now();
		chrono::nanoseconds duration {};
		chrono::nanoseconds stop(){
			duration = chrono::steady_clock::now() - start;
			return duration;
		}

		template<typename Unit = chrono::microseconds>
		void stop_and_print_result(std::string_view msg, unsigned parts){
			duration = chrono::steady_clock::now() - start;
			println("{}: {}", msg, chrono::duration_cast<Unit> (duration / parts));
		}
	};

	template<typename Unit = chrono::microseconds, typename Fun>
	void benchmark(std::string_view msg, unsigned times, Fun fun){
		Timer t;
		for(unsigned i = 0; i < times; i++){
			fun();
		}
		t.stop_and_print_result<Unit>(msg, times);
	}


	class App {
		Context context;

		wgpu::Buffer stagingBuffer;

		wgpu::ShaderModule computeShader;
		wgpu::BindGroupLayout bindGroupLayout;
		wgpu::ComputePipeline computePipeline;

		std::vector<uint8_t> image;
		
	public:
		App() :
			context{false}
		{
			//constexpr unsigned keysPerThread = 16;
			//constexpr unsigned bitsPerPass = 4;

			
			std::vector<uint32_t> keys = generate_data(numKeys);
			std::vector<uint32_t> values = generate_data(numKeys);
			
			//ranges::sort(values);

			RadixSort radixSort(context, kernelsPath);
			PrefixSum prefixSum(context, kernelsPath);
			auto tmpSize = radixSort.getScratchDataSize(static_cast<uint32_t>(keys.size()));

			using Usage = wgpu::BufferUsage;
			Buffer keysBuffer(context, byteSize(keys), Usage::CopyDst | Usage::Storage, false, asBytes(keys));
			Buffer valuesBuffer(context, byteSize(values), Usage::CopyDst | Usage::Storage, false, asBytes(values));
			Buffer tmpBuffer(context, tmpSize, Usage::Storage);//todo remove | Usage::CopySrc
			Buffer sortedKeysBuffer(context, byteSize(keys), Usage::Storage | Usage::CopySrc);
			Buffer sortedValuesBuffer(context, byteSize(values), Usage::Storage | Usage::CopySrc);
			context.flushCommandsAndWait();

			std::vector<uint32_t> smallKeys = generate_data(1024);
			Buffer smallKeysBuffer(context, 1024*sizeof(uint32_t), Usage::CopyDst | Usage::Storage | Usage::CopySrc);
			Buffer smallValuesBuffer(context, 1024*sizeof(uint32_t), Usage::CopyDst | Usage::Storage);
			Buffer smallTmpBuffer(context, radixSort.getScratchDataSize(1024u), Usage::Storage);
			Buffer smallSortedKeysBuffer(context, 1024*sizeof(uint32_t), Usage::Storage | Usage::CopySrc);
			Buffer smallSortedValuesBuffer(context, 1024*sizeof(uint32_t), Usage::Storage | Usage::CopySrc);

			auto prefixSumScratchSize = prefixSum.getScratchDataSize(static_cast<uint32_t>(keys.size()));
			/*
			for(int i = 0; i < 4; i++) { 
				prefixSum.run(keysBuffer, sortedKeysBuffer, tmpBuffer.subrange(0, prefixSumScratchSize), 
					static_cast<uint32_t>(keys.size()));
				context.flushCommandsAndWait();
			}
			*//*
		    benchmark("PrefixSum", 10, [&](){
				prefixSum.run(keysBuffer, sortedKeysBuffer, tmpBuffer.subrange(0, prefixSumScratchSize), 
					static_cast<uint32_t>(keys.size()));
				context.flushCommandsAndWait();
			});
			
			std::vector<uint32_t> prefixSumCPU(numKeys);
			prefixSumCPU[0] = 0;
			benchmark("PrefixSumCPU", 1, [&](){
				for (size_t i = 1; i < numKeys; i++){
					prefixSumCPU[i] = prefixSumCPU[i - 1] + keys[i];
				}
			});
			println();*/
			
			std::vector<uint32_t> gpuSortedKeys(keys.size());
			chrono::nanoseconds duration{};
			for(int i = 0; i < 10; i++) {
				keysBuffer.setData(std::as_bytes(std::span{keys}));
				smallKeysBuffer.setData(std::as_bytes(std::span{smallKeys}));
				context.flushCommandsAndWait();

				Timer rs_timer;
				radixSort.run(keysBuffer, sortedKeysBuffer, valuesBuffer, sortedValuesBuffer, tmpBuffer, static_cast<uint32_t>(keys.size()));
				context.flushCommandsAndWait();
				duration += rs_timer.stop();

				sortedKeysBuffer.getData(std::as_writable_bytes(std::span{ gpuSortedKeys }));
				smallKeysBuffer.getData(std::as_writable_bytes(std::span{smallKeys}));
				context.flushCommandsAndWait();
				radixSort.run(smallKeysBuffer, smallSortedKeysBuffer, smallValuesBuffer, smallSortedValuesBuffer, smallTmpBuffer, 1024);
				context.flushCommandsAndWait();
			}
			println("{}: {}", "RadixSort", chrono::duration_cast<chrono::microseconds> (duration / 10));

			std::vector<std::pair<uint32_t, uint32_t>> cpuSorted(numKeys);
			for (unsigned i = 0; i < numKeys; i++) {
				cpuSorted[i] = std::pair{ keys[i], values[i] };
			}
			
			benchmark<chrono::microseconds>("CPU sort", 1, [&](){
				ranges::stable_sort(cpuSorted, {}, [](auto pair) { return pair.first; });
			});
			

			std::vector<std::pair<uint32_t, uint32_t>> pairs(numKeys);
			
			std::vector<uint32_t> gpuSortedValues(keys.size());
			context.flushCommandsAndWait();
			sortedKeysBuffer.getData(std::as_writable_bytes(std::span{ gpuSortedKeys }));
			context.flushCommandsAndWait();
			sortedValuesBuffer.getData(std::as_writable_bytes(std::span{ gpuSortedValues }));
			context.flushCommandsAndWait();
			
			std::cout << std::hex;
			for (int i = 0; i < keys.size(); i++) {
				//std::cout << gpuSortedKeys[i] << "     " << cpuSorted[i].first << '\n';
				if (std::pair{ gpuSortedKeys[i], gpuSortedValues[i] } != cpuSorted[i]) {
					println("cpu:{} {}, gpu: {} {}",
						cpuSorted[i].first, cpuSorted[i].second,
						gpuSortedKeys[i], gpuSortedValues[i]);
					//break;
				}
			}
			
			/*
			image.resize(long long{ context.imageSize.width } * context.imageSize.height * 4, 255);

			wgpu::BufferDescriptor bufferDesc1{
				.usage = wgpu::BufferUsage::MapWrite | wgpu::BufferUsage::CopySrc,
				.size = 960*540*4,
				.mappedAtCreation = true
			};


			stagingBuffer = context.device.CreateBuffer(&bufferDesc1);
			//auto buffer = stagingBuffer.GetMappedRange();
			
			int width, height;
			auto* stbiImage = stbi_load("./picture.jpg", &width, &height, nullptr, 4);
			assert(stbiImage);

			Buffer buffer(context, width * height * 4, wgpu::BufferUsage::CopySrc | wgpu::BufferUsage::CopyDst, false,
				reinterpret_cast<const std::byte*>(stbiImage));
			stbi_image_free(stbiImage);
			buffer.copyToRenderImage();
			context.flushCommands();
			*/
			println("App's constructor ended.");
		}
		void run() {
			while (!context.shouldClose()) {
				context.pollEvents();

				context.drawRenderTexture();
				context.flushCommands();
			}
			print("App end");
		}
	};

}


int main() {
	try {
		::App app;
		//app.run();
	}
	catch (std::exception e) {
		std::cout << e.what() << std::endl;
		return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}
