#include "radix.h"
#include "bindGroups.h"
#include "buffer.h"
#include "common.h"

#include<array>
#include<cstring>

constexpr uint32_t keysPerThread = 8u;
constexpr uint32_t groupSize = 256u;
constexpr uint32_t keysPerGroup = groupSize * keysPerThread;

constexpr uint32_t numBitsPerPass = 4;
constexpr uint32_t histogramNumBins = (1 << numBitsPerPass);
constexpr unsigned passes = 32u / numBitsPerPass;


namespace {
    enum class ScanType
    {
        Block,
        TwoLevel,
        ThreeLevel
    };

    ScanType getScanType(std::uint32_t numKeys)
    {
        if (numKeys <= keysPerGroup)
        {
            return ScanType::Block;
        }
        else
        {
            numKeys = ceilDivide(numKeys, keysPerGroup);
            if (numKeys <= keysPerGroup)
            {
                return ScanType::TwoLevel;
            }
            else
            {
                return ScanType::ThreeLevel;
            }
        }
    }

    constexpr uint32_t calculateNumGroupHistogramElements(uint32_t numKeys) {
        auto num_histograms = ceilDivide(numKeys, keysPerGroup);
        return num_histograms * histogramNumBins;
    };

    struct PushConstantUniform {
        uint32_t num_keys;
        uint32_t bit_shift;
    };

    void setPushConstantsBuffer(Buffer& result, uint32_t numKeys) {

        std::array<std::byte, bufferSubrangeAllignment * passes> pushConstantCpuBuffer{};
        
        for (auto pass = 0u; pass < passes; pass++) {
            unsigned bitshift = pass * numBitsPerPass;
            PushConstantUniform pushConstants{ numKeys, bitshift };
            memcpy(&pushConstantCpuBuffer[size_t{ pass } * bufferSubrangeAllignment], 
                &pushConstants, sizeof(pushConstants));
        }

        result.setData(pushConstantCpuBuffer);

        /*
        auto& cpuToGpu = Buffer::stagingBuffers->cpuToGpu;
        
        context.flushCommands();

        cpuToGpu.reserve(bufferSubrangeAllignment * passes);
        cpuToGpu.mapAsync();
        auto* stagingMemory = cpuToGpu.getMapped();

        for (unsigned pass = 0; pass < passes; pass++) {
            unsigned bitshift = pass * numBitsPerPass;
            PushConstantUniform pushConstants{ numKeys, bitshift };
            memcpy(&stagingMemory[bufferSubrangeAllignment * pass], &pushConstants, sizeof(pushConstants));
        }

	    cpuToGpu.unmap();
        cpuToGpu.copy_to(result);
        */
    }
}


RadixSort::RadixSort(const Context& context, const std::filesystem::path& kernelsPath) :
	_context{ context },
	_groupHistograms(context, kernelsPath / "radix_sort_group_histograms.wgsl"),
	_scatterKeys(context, kernelsPath / "radix_sort_scatter_keys.wgsl"),
	_scan(context, kernelsPath),
	_pushConstantsBuffer(context, passes * bufferSubrangeAllignment, wgpu::BufferUsage::Uniform | wgpu::BufferUsage::CopyDst)
{
	_groupHistograms.setBindGroup(0);
	_scatterKeys.setBindGroup(0);
}




void RadixSort::run(BufferSubrange inputKeys, BufferSubrange outputKeys, 
    BufferSubrange inputValues, BufferSubrange outputValues, 
    BufferSubrange scratch, uint32_t numKeys) 
{
    auto numHistogramElements = calculateNumGroupHistogramElements(numKeys);
    
    assert(scratch.size() == getScratchDataSize(numKeys));
    auto [scanScratch, offset1] = scratch.subrange(0, _scan.getScratchDataSize(numHistogramElements));
    auto [tmpKeys, offset2] = scratch.subrange(offset1, numKeys * sizeof(uint32_t));
    auto [tmpValues, offset3] = scratch.subrange(offset2, numKeys * sizeof(uint32_t));
    auto [histograms, end] = scratch.subrange(offset3, numHistogramElements * sizeof(uint32_t));
    assert(end <= scratch.size());

    if (_lastNumKeys != numKeys){
        _lastNumKeys = numKeys;
        setPushConstantsBuffer(_pushConstantsBuffer, numKeys);
    }

    for (unsigned pass = 0; pass < passes; pass++) {
        unsigned bitshift = pass * numBitsPerPass;
        
        _groupHistograms.setBindGroup(2, 
            _pushConstantsBuffer.subrange(size_t{ pass } * bufferSubrangeAllignment, sizeof(PushConstantUniform)));
        
        bool oddRound = bitshift & ((numBitsPerPass << 1) - 1);

        if (bitshift == 0) { //start
            _groupHistograms.setBindGroup(1, inputKeys, histograms); 
        }
        else if (oddRound) { //ping
            _groupHistograms.setBindGroup(1, tmpKeys, histograms);
        }
        else { //pong
            _groupHistograms.setBindGroup(1, outputKeys, histograms);
        }
        _groupHistograms.dispatch(ceilDivide(numKeys, keysPerGroup));

        _scan.run(histograms, histograms, scanScratch, numHistogramElements);


        _scatterKeys.setBindGroup(2, 
            _pushConstantsBuffer.subrange(size_t{ pass } * bufferSubrangeAllignment, sizeof(PushConstantUniform)));
        
        if (bitshift == 0) { //start
            _scatterKeys.setBindGroup(1, inputKeys, histograms, tmpKeys, inputValues, tmpValues); 
        }
        else if (oddRound) { //ping
            _scatterKeys.setBindGroup(1, tmpKeys, histograms, outputKeys, tmpValues, outputValues);
        }
        else { //pong
            _scatterKeys.setBindGroup(1, outputKeys, histograms, tmpKeys, outputValues, tmpValues);
        }
        _scatterKeys.dispatch(ceilDivide(numKeys, keysPerGroup));
    }
}

size_t RadixSort::getScratchDataSize(uint32_t numKeys) const {
    auto tmpKeysSize = roundUp(numKeys * sizeof(uint32_t), bufferSubrangeAllignment);
    auto tmpValuesSize = tmpKeysSize;
    auto numHistogramElements = calculateNumGroupHistogramElements(numKeys);
    auto histSize = roundUp(numHistogramElements * sizeof(uint32_t), bufferSubrangeAllignment);
    auto scanScratchSize = _scan.getScratchDataSize(numHistogramElements);
    assert(scanScratchSize % bufferSubrangeAllignment == 0);
    return scanScratchSize + tmpKeysSize + tmpValuesSize + histSize;
}