#include "scan.h"
#include "bindGroups.h"
#include "buffer.h"
#include "common.h"

#include <array>

constexpr uint32_t keysPerThread = 2u;
constexpr uint32_t groupSize = 256u;
constexpr uint32_t keysPerGroup = keysPerThread * groupSize;

namespace {
    enum class ScanType {
        Block,
        TwoLevel,
        ThreeLevel
    };

    ScanType getScanType(std::uint32_t numKeys)
    {
        if (numKeys <= keysPerGroup)
        {
            return ScanType::Block;
        }
        else
        {
            numKeys = ceilDivide(numKeys, keysPerGroup);
            if (numKeys <= keysPerGroup)
            {
                return ScanType::TwoLevel;
            }
            else
            {
                return ScanType::ThreeLevel;
            }
        }
    }

    struct ScanUniform {
        uint32_t numKeys;
        uint32_t addSum;
    };

    enum class pushConstantIndex{ 
        reduceLevel1, 
        reduceLevel2, 
        scanLevel0, 
        scanLevel1, 
        scanLevel2
    };

    unsigned to_buffer_pos(pushConstantIndex index){
        return static_cast<int>(index) * bufferSubrangeAllignment;
    }
}



void setPushConstantsBuffer(Buffer& result, uint32_t numKeys, ScanType scanType) {
    assert(scanType == getScanType(numKeys));
    const uint32_t numGroupsLevel1 = ceilDivide(numKeys, keysPerGroup);
    const uint32_t numGroupsLevel2 = ceilDivide(numGroupsLevel1, keysPerGroup);

    std::array<std::byte, 5 * bufferSubrangeAllignment> pushConstantBufferCPU{};
    
    using enum pushConstantIndex;
    auto setConstant = [&pushConstantBufferCPU](pushConstantIndex index, auto val){ 
        memcpy(&pushConstantBufferCPU[to_buffer_pos(index)], &val, sizeof(val)); 
    };

    setConstant(reduceLevel1, numKeys);
    setConstant(reduceLevel2, numGroupsLevel1);
    
    ScanUniform scanUniLevel2{ numGroupsLevel2, 0u };
    setConstant(scanLevel2, scanUniLevel2);
    
    ScanUniform scanUniLevel1{ numGroupsLevel1, (scanType != ScanType::TwoLevel) ? 1u : 0u };
    setConstant(scanLevel1, scanUniLevel1);
    
    ScanUniform scanUniLevel0{ numKeys, (scanType != ScanType::Block) ? 1u : 0u };
    setConstant(scanLevel0, scanUniLevel0);
    
    result.setData(pushConstantBufferCPU);
}

void PrefixSum::run(BufferSubrange input, BufferSubrange output, BufferSubrange scratch, uint32_t numKeys){
    assert(scratch.size() == getScratchDataSize(numKeys));
    const auto partSums0Size = ceilDivide(numKeys, keysPerGroup);
    auto [partSums0, offset1] = scratch.subrange(0, partSums0Size * sizeof(uint32_t));
    auto [partSums1, end] = scratch.subrange(offset1, ceilDivide(partSums0Size, keysPerGroup) * sizeof(uint32_t));
    assert(end <= scratch.size());

    const auto scanType = getScanType(numKeys);
    

    if (numKeys != _lastNumKeys){
        _lastNumKeys = numKeys;
        setPushConstantsBuffer(_pushConstants, numKeys, scanType);
    }

    using enum pushConstantIndex;
    const uint32_t numGroupsLevel1 = ceilDivide(numKeys, keysPerGroup);
    if (scanType != ScanType::Block) {
        _reduce.setBindGroup(1, input, partSums0);
        _reduce.setBindGroup(2, _pushConstants.subrange(to_buffer_pos(reduceLevel1), sizeof(uint32_t)));
        _reduce.dispatch(numGroupsLevel1);
        
        const uint32_t numGroupsLevel2 = ceilDivide(numGroupsLevel1, keysPerGroup);
        if (scanType != ScanType::TwoLevel) {
            _reduce.setBindGroup(1, partSums0, partSums1);
            _reduce.setBindGroup(2, _pushConstants.subrange(to_buffer_pos(reduceLevel2), sizeof(uint32_t)));
            _reduce.dispatch(numGroupsLevel2);

            _scanAdd.setBindGroup(1, partSums1, partSums1, partSums1);
            _scanAdd.setBindGroup(2, _pushConstants.subrange(to_buffer_pos(scanLevel2), sizeof(ScanUniform)));
            assert(ceilDivide(numGroupsLevel2, keysPerGroup) == 1);
            _scanAdd.dispatch(1);
        }

        _scanAdd.setBindGroup(1, partSums0, partSums1, partSums0);
        _scanAdd.setBindGroup(2, _pushConstants.subrange(to_buffer_pos(scanLevel1), sizeof(ScanUniform)));
        _scanAdd.dispatch(numGroupsLevel2);
    }

    _scanAdd.setBindGroup(1, input, partSums0, output);
    _scanAdd.setBindGroup(2, _pushConstants.subrange(to_buffer_pos(scanLevel0), sizeof(ScanUniform)));
    _scanAdd.dispatch(numGroupsLevel1);
    _context.flushCommands();
}

size_t PrefixSum::getScratchDataSize(uint32_t num_keys) const {
	const auto partSums0Count = ceilDivide(num_keys, keysPerGroup);
	const size_t partSums0Size = roundUp(partSums0Count * sizeof(uint32_t), bufferSubrangeAllignment);
    
    size_t partSums1Size = bufferSubrangeAllignment;
    if (partSums0Count > keysPerGroup)
    {
        partSums1Size = roundUp(ceilDivide(partSums0Count, keysPerGroup) * sizeof(uint32_t), bufferSubrangeAllignment);
    }
    return partSums0Size + partSums1Size;
}
